from __future__ import annotations

import base64
import io
import mimetypes
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncIterator, Collection, ContextManager, Dict, List, Optional, Tuple

import anthropic
import httpx
import openai
import tiktoken
from anthropic import AsyncAnthropic
from anthropic.types import Message, MessageParam, MessageStreamEvent, TextBlockParam
from openai import AsyncAzureOpenAI, AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from PIL import Image
from typing_extensions import override

from spice.errors import APIConnectionError, APIError, AuthenticationError, ImageError, InvalidModelError
from spice.spice_message import VALID_MIMETYPES, SpiceMessage

if TYPE_CHECKING:
    from spice.models import Model
    from spice.spice import SpiceCallArgs


class WrappedClient(ABC):
    @abstractmethod
    async def get_chat_completion_or_stream(
        self, call_args: SpiceCallArgs
    ) -> ChatCompletion | AsyncIterator[ChatCompletionChunk] | Message | AsyncIterator[MessageStreamEvent]: ...

    @abstractmethod
    def process_chunk(self, chunk) -> tuple[Optional[str], Optional[int], Optional[int]]: ...

    @abstractmethod
    def extract_text_and_tokens(self, chat_completion) -> tuple[str, int, int]: ...

    @abstractmethod
    def catch_and_convert_errors(self) -> ContextManager[None]: ...

    @abstractmethod
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        """
        Returns the number of tokens used by a prompt if it was sent to an API for a chat completion.
        """

    @abstractmethod
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        """
        Calculates the tokens in this message. Will not be accurate for a prompt.
        Use count_messages_tokens to get the exact amount of tokens for a prompt.
        If full_message is true, will include the extra 4 tokens used in a chat completion by this message
        if this message is part of a prompt. Do not full_message to true for a response.
        """

    @abstractmethod
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]: ...

    @abstractmethod
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]: ...

    @abstractmethod
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]: ...


class WrappedOpenAIClient(WrappedClient):
    def __init__(self, key, base_url=None):
        self._sync_client = OpenAI(api_key=key, base_url=base_url)
        self._client = AsyncOpenAI(api_key=key, base_url=base_url)

    @override
    async def get_chat_completion_or_stream(self, call_args: SpiceCallArgs):
        # WrappedOpenAIClient can be used with a proxy to a non openai llm, which may not support response_format
        maybe_response_format_kwargs: Dict[str, Any] = (
            {"response_format": call_args.response_format}
            if call_args.response_format is not None and "type" in call_args.response_format
            else {}
        )

        # GPT-4-vision has low default max_tokens
        if call_args.max_tokens is None and "gpt-4" in call_args.model and "vision-preview" in call_args.model:
            max_tokens = 4096
        else:
            max_tokens = call_args.max_tokens

        return await self._client.chat.completions.create(
            model=call_args.model,
            messages=list(call_args.messages),
            stream=call_args.stream,
            temperature=call_args.temperature,
            max_tokens=max_tokens,
            **maybe_response_format_kwargs,
        )

    @override
    def process_chunk(self, chunk):
        content = chunk.choices[0].delta.content
        return content, None, None

    @override
    def extract_text_and_tokens(self, chat_completion):
        return (
            chat_completion.choices[0].message.content,
            chat_completion.usage.prompt_tokens,
            chat_completion.usage.completion_tokens,
        )

    @override
    @contextmanager
    def catch_and_convert_errors(self):
        # TODO: Do we catch all errors? I think we should catch APIStatusError
        try:
            yield
        except openai.APIConnectionError as e:
            raise APIConnectionError(f"OpenAI Connection Error: {e.message}") from e
        except openai.AuthenticationError as e:
            raise AuthenticationError(f"OpenAI Authentication Error: {e.message}") from e
        except openai.APIStatusError as e:
            raise APIError(f"OpenAI Status Error: {e.message}") from e

    def _get_encoding_for_model(self, model: Model | str) -> tiktoken.Encoding:
        from spice.models import Model

        if isinstance(model, Model):
            model = model.name
        try:
            # OpenAI fine-tuned models are named `ft:<base model>:<name>:<id>`. If tiktoken
            # can't match the full string, it tries to match on startswith, e.g. 'gpt-4'
            _model = model.split(":")[1] if model.startswith("ft:") else model
            return tiktoken.encoding_for_model(_model)
        except KeyError:
            return tiktoken.get_encoding("cl100k_base")

    @override
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        """
        Adapted from https://platform.openai.com/docs/guides/text-generation/managing-tokens
        """

        encoding = self._get_encoding_for_model(model)

        num_tokens = 0
        for message in messages:
            # every message follows <|start|>{role/name}\n{content}<|end|>\n
            # this has 5 tokens (start token, role, \n, end token, \n), but we count the role token later
            num_tokens += 4
            for key, value in message.items():
                if isinstance(value, list) and key == "content":
                    for entry in value:
                        if entry["type"] == "text":
                            num_tokens += len(encoding.encode(entry["text"]))
                        if entry["type"] == "image_url":
                            image_base64: str = entry["image_url"]["url"].split(",")[1]
                            image_bytes: bytes = base64.b64decode(image_base64)
                            image = Image.open(io.BytesIO(image_bytes))
                            size = image.size
                            # As described here: https://platform.openai.com/docs/guides/vision/calculating-costs
                            scale = min(1, 2048 / max(size))
                            size = (int(size[0] * scale), int(size[1] * scale))
                            scale = min(1, 768 / min(size))
                            size = (int(size[0] * scale), int(size[1] * scale))
                            num_tokens += 85 + 170 * ((size[0] + 511) // 512) * ((size[1] + 511) // 512)
                elif isinstance(value, str):
                    num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens -= 1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <|start|>assistant
        return num_tokens

    @override
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        encoding = self._get_encoding_for_model(model)
        return len(encoding.encode(message, disallowed_special=())) + (4 if full_message else 0)

    @override
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]:
        embeddings = (await self._client.embeddings.create(input=input_texts, model=model)).data
        sorted_embeddings = sorted(embeddings, key=lambda e: e.index)
        return [result.embedding for result in sorted_embeddings]

    @override
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]:
        embeddings = self._sync_client.embeddings.create(input=input_texts, model=model).data
        sorted_embeddings = sorted(embeddings, key=lambda e: e.index)
        return [result.embedding for result in sorted_embeddings]

    @override
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]:
        audio_file = open(audio_path, "rb")
        transcript = await self._client.audio.transcriptions.create(
            model=model, file=audio_file, response_format="verbose_json"
        )
        return (transcript.text, transcript.duration)  # pyright: ignore


class WrappedAzureClient(WrappedOpenAIClient):
    def __init__(self, key, endpoint):
        self._client = AsyncAzureOpenAI(
            api_key=key,
            api_version="2023-12-01-preview",
            azure_endpoint=endpoint,
        )


class WrappedAnthropicClient(WrappedClient):
    def __init__(self, key):
        self._client = AsyncAnthropic(api_key=key)

    def _convert_messages(
        self, messages: Collection[SpiceMessage], add_json_brace: bool
    ) -> Tuple[str, List[MessageParam]]:
        # Anthropic handles both images and system messages different from OpenAI, only allows alternating user / assistant messages,
        # and doesn't support tools / function calling (still in beta, and doesn't support streaming)

        system = ""
        converted_messages: List[MessageParam] = []
        start = True
        cur_role = ""
        for message in messages:
            if message["role"] == "system" and start:
                if system:
                    system += "\n\n"
                system += message["content"]
                continue

            # First message must be user, and text content cannot be empty / whitespace only
            if start and message["role"] != "user":
                cur_role = "user"
                converted_messages.append({"role": "user", "content": [{"type": "text", "text": "-"}]})

            start = False

            # Anthropic messages can either be a string or list of objects; since user messages can have images, they should always be a list objects.
            # Assistant messages should always be strings to keep them simple.
            match message["role"]:
                case "system":
                    if cur_role == "user":
                        message_object: TextBlockParam = {"type": "text", "text": f"\n\nSystem:\n{message['content']}"}
                        converted_messages[-1]["content"].append(message_object)  # pyright: ignore
                    else:
                        cur_role = "user"
                        message_object: TextBlockParam = {"type": "text", "text": f"System:\n{message['content']}"}
                        converted_messages.append({"role": "user", "content": [message_object]})
                case "assistant":
                    content = message.get("content", "")
                    if content is None:
                        content = ""
                    if cur_role == "assistant":
                        converted_messages[-1]["content"] += f"\n\n{content}"  # pyright: ignore
                    else:
                        cur_role = "assistant"
                        converted_messages.append({"role": "assistant", "content": content})
                case "user":
                    if isinstance(message["content"], str):
                        if cur_role == "user":
                            converted_messages[-1]["content"].append(  # pyright: ignore
                                {"type": "text", "text": f"\n\n{message['content']}"}
                            )
                        else:
                            converted_messages.append(
                                {"role": "user", "content": [{"type": "text", "text": f"{message['content']}"}]}
                            )
                    else:
                        first = cur_role == "user"
                        content = []
                        for sub_content in message["content"]:
                            if sub_content["type"] == "text":
                                content.append(
                                    {"type": "text", "text": ("\n\n" if first else "") + sub_content["text"]}
                                )
                            else:
                                # This can either be base64 encoded data or a url; Anthropic only accepts base64 encoded data
                                image = sub_content["image_url"]["url"]
                                if image.startswith("http"):
                                    try:
                                        response = httpx.get(image)
                                    except:
                                        raise ImageError(f"Error fetching image {image}.")

                                    media_type = response.headers.get("content-type", mimetypes.guess_type(image)[0])
                                    image = base64.b64encode(response.content).decode("utf-8")
                                else:
                                    media_type = image.split(";", maxsplit=1)[0].replace("data:", "")
                                    image = image.split(";base64,", maxsplit=1)[1]

                                if media_type not in VALID_MIMETYPES:
                                    raise ImageError(
                                        f"Invalid image at {image}: Image must be a png, jpg, gif, or webp image."
                                    )

                                content.append(
                                    {
                                        "type": "image",
                                        "source": {"type": "base64", "data": image, "media_type": media_type},
                                    }
                                )

                            first = False
                        if cur_role == "user":
                            converted_messages[-1]["content"].extend(content)  # pyright: ignore
                        else:
                            converted_messages.append({"role": "user", "content": content})

                    cur_role = "user"
                case "tool":
                    # Right now anthropic tool use is in beta and doesn't support some things like streaming, but once it releases we can modify this.
                    pass
                case "function":
                    # Deprecated, nobody should use this
                    pass

        if add_json_brace and converted_messages:
            if converted_messages[-1]["role"] == "assistant" and not converted_messages[-1]["content"]:
                converted_messages[-1]["content"] += "{"  # pyright: ignore
            elif converted_messages[-1]["role"] == "user":
                converted_messages.append({"role": "assistant", "content": "{"})

        return system, converted_messages

    @override
    async def get_chat_completion_or_stream(self, call_args: SpiceCallArgs):
        add_json_brace = (
            call_args.response_format is not None and call_args.response_format.get("type", "text") == "json_object"
        )

        # max_tokens is required by anthropic api
        if call_args.max_tokens is None:
            max_tokens = 4096
        else:
            max_tokens = call_args.max_tokens

        # temperature is optional but can't be None
        maybe_temperature_kwargs: dict[str, Any] = (
            {"temperature": call_args.temperature} if call_args.temperature is not None else {}
        )

        system, converted_messages = self._convert_messages(call_args.messages, add_json_brace)

        return await self._client.messages.create(
            model=call_args.model,
            system=system,
            messages=converted_messages,
            stream=call_args.stream,
            max_tokens=max_tokens,
            **maybe_temperature_kwargs,
        )

    @override
    def process_chunk(self, chunk):
        content = None
        input_tokens = None
        output_tokens = None
        if chunk.type == "content_block_delta":
            content = chunk.delta.text
        elif chunk.type == "message_start":
            input_tokens = chunk.message.usage.input_tokens
        elif chunk.type == "message_delta":
            output_tokens = chunk.usage.output_tokens
        return content, input_tokens, output_tokens

    @override
    def extract_text_and_tokens(self, chat_completion):
        return (
            chat_completion.content[0].text,
            chat_completion.usage.input_tokens,
            chat_completion.usage.output_tokens,
        )

    @override
    @contextmanager
    def catch_and_convert_errors(self):
        try:
            yield
        except anthropic.APIConnectionError as e:
            raise APIConnectionError(f"Anthropic Connection Error: {e.message}") from e
        except anthropic.AuthenticationError as e:
            raise AuthenticationError(f"Anthropic Authentication Error: {e.message}") from e
        except anthropic.APIStatusError as e:
            raise APIError(f"Anthropic Status Error: {e.message}") from e

    # Anthropic doesn't give us a way to count tokens, so we just use OpenAI's token counting functions and multiply by a pre-determined multiplier
    class _FakeWrappedOpenAIClient(WrappedOpenAIClient):
        def __init__(self):
            pass

    _fake_openai_client = _FakeWrappedOpenAIClient()
    _anthropic_token_multiplier = 1.25

    @override
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        from spice.models import GPT_35_TURBO_0125

        return int(
            self._fake_openai_client.count_messages_tokens(messages, GPT_35_TURBO_0125)
            * self._anthropic_token_multiplier
        )

    @override
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        from spice.models import GPT_35_TURBO_0125

        return int(
            self._fake_openai_client.count_string_tokens(message, GPT_35_TURBO_0125, full_message)
            * self._anthropic_token_multiplier
        )

    @override
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]:
        raise InvalidModelError()



from contextlib import contextmanager
from pathlib import Path
from typing import AsyncIterator, Collection, Iterator, List, Optional

from openai.types.chat import ChatCompletion, ChatCompletionChunk, ChatCompletionMessage
from openai.types.chat.chat_completion import Choice as CompletionChoice
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDelta
from typing_extensions import override

from spice.errors import InvalidModelError
from spice.models import Model
from spice.spice import SpiceCallArgs
from spice.spice_message import SpiceMessage
from spice.wrapped_clients import WrappedClient


async def convert_string_to_asynciter(
    input_str: str,
    chunk_size: int = 5,
) -> AsyncIterator[ChatCompletionChunk]:
    for i in range(0, len(input_str), chunk_size):
        yield ChatCompletionChunk(
            id="",
            choices=[
                Choice(
                    index=0,
                    delta=ChoiceDelta(
                        content=input_str[i : i + chunk_size],
                    ),
                )
            ],
            created=0,
            model="",
            object="chat.completion.chunk",
        )


class WrappedTestClient(WrappedClient):
    """
    A wrapped client that can be used in tests. Accepts what it should respond with in its constructor.
    """

    def __init__(self, response: str | Iterator[str]):
        if isinstance(response, str):
            self.response = iter(response)
        else:
            self.response = response

    @override
    async def get_chat_completion_or_stream(
        self, call_args: SpiceCallArgs
    ) -> ChatCompletion | AsyncIterator[ChatCompletionChunk]:
        response = next(self.response)

        if call_args.stream:
            return convert_string_to_asynciter(response)
        else:
            return ChatCompletion(
                id="",
                choices=[
                    CompletionChoice(
                        finish_reason="stop",
                        index=0,
                        message=ChatCompletionMessage(
                            content=response,
                            role="assistant",
                        ),
                    )
                ],
                created=0,
                model="",
                object="chat.completion",
            )

    @override
    @contextmanager
    def catch_and_convert_errors(self):
        yield

    @override
    def count_messages_tokens(self, messages: Collection[SpiceMessage], model: Model | str) -> int:
        return 0

    @override
    def count_string_tokens(self, message: str, model: Model | str, full_message: bool) -> int:
        return 0

    @override
    def process_chunk(self, chunk: ChatCompletionChunk) -> tuple[str, Optional[int], Optional[int]]:
        return chunk.choices[0].delta.content or "", None, None

    @override
    def extract_text_and_tokens(self, chat_completion) -> tuple[str, int, int]:
        return (chat_completion.content[0].text, 0, 0)

    @override
    async def get_embeddings(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    def get_embeddings_sync(self, input_texts: List[str], model: str) -> List[List[float]]:
        raise InvalidModelError()

    @override
    async def get_transcription(self, audio_path: Path, model: str) -> tuple[str, float]:
        raise InvalidModelError()



from __future__ import annotations

import base64
import mimetypes
from collections import UserList
from json import JSONEncoder
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, TypedDict

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)

from spice.errors import ImageError

if TYPE_CHECKING:
    from spice.spice import Spice

SpiceMessage = ChatCompletionMessageParam
"""The message format Spice uses and accepts. Exactly the same as OpenAI's message format."""

VALID_MIMETYPES = ["image/jpeg", "image/png", "image/gif", "image/webp"]


def user_message(content: str) -> ChatCompletionUserMessageParam:
    """Creates a user message with the given content."""
    return {"role": "user", "content": content}


def system_message(content: str) -> ChatCompletionSystemMessageParam:
    """Creates a system message with the given content."""
    return {"role": "system", "content": content}


def assistant_message(content: str) -> ChatCompletionAssistantMessageParam:
    """Creates an assistant message with the given content."""
    return {"role": "assistant", "content": content}


def image_bytes_message(image_bytes: bytes, media_type: str) -> ChatCompletionUserMessageParam:
    """Creates a user message containing the given image bytes."""
    if media_type not in VALID_MIMETYPES:
        raise ImageError(f"Invalid image type {media_type}: Image must be a png, jpg, gif, or webp image.")

    image = base64.b64encode(image_bytes).decode("utf-8")
    return {
        "role": "user",
        "content": [{"type": "image_url", "image_url": {"url": f"data:{media_type};base64,{image}"}}],
    }


def file_image_message(file_path: Path | str) -> ChatCompletionUserMessageParam:
    """Creates a user message with the image at the given path. The image must be a png, jpg, gif, or webp image."""

    file_path = Path(file_path).expanduser().resolve()
    if not file_path.exists():
        raise ImageError(f"Invalid image at {file_path}: file does not exist.")

    media_type = mimetypes.guess_type(file_path)[0]
    if media_type not in VALID_MIMETYPES:
        raise ImageError(f"Invalid image at {file_path}: Image must be a png, jpg, gif, or webp image.")

    with file_path.open("rb") as file:
        image_bytes = file.read()

    return image_bytes_message(image_bytes, media_type)


def http_image_message(url: str) -> ChatCompletionUserMessageParam:
    """Creates a user message with an image from the given url."""
    if not (url.startswith("http://") or url.startswith("https://")):
        raise ImageError(f"Invalid image URL {url}: Must be http or https protocol.")

    return {"role": "user", "content": [{"type": "image_url", "image_url": {"url": url}}]}


class MessagesEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, SpiceMessages):
            new_data = []
            for message in o.data:
                # Shallow copy is ok since we won't be changing anything other than prompt_metadata
                new_message = message.copy()
                if hasattr(message, "prompt_metadata"):
                    new_message["prompt_metadata"] = message.prompt_metadata  # pyright: ignore
                new_data.append(new_message)
            return new_data
        else:
            return super().default(o)


class PromptMetadata(TypedDict):
    name: str
    content: str
    context: Dict[str, Any]


# Regular dicts can't have arbitrary attributes attached to them
class _MetadataDict(dict):
    prompt_metadata: PromptMetadata


class SpiceMessages(UserList[SpiceMessage]):
    """
    A light wrapper around the messages list simplifing adding new messages.
    """

    def __init__(self, client: Spice, initlist: Optional[Iterable[SpiceMessage]] = None):
        self._client = client
        super().__init__(initlist)

    def add_user_message(self, content: str):
        """Appends a user message with the given content."""
        self.data.append(user_message(content))

    def add_system_message(self, content: str):
        """Appends a system message with the given content."""
        self.data.append(system_message(content))

    def add_assistant_message(self, content: str):
        """Appends an assistant message with the given content."""
        self.data.append(assistant_message(content))

    def add_image_bytes_message(self, image_bytes: bytes, media_type: str):
        """Appends a user message with the given image bytes."""
        self.data.append(image_bytes_message(image_bytes, media_type))

    def add_file_image_message(self, file_path: Path | str):
        """Appends a user message with the image from the given file. The image must be a png, jpg, gif, or webp image."""
        self.data.append(file_image_message(file_path))

    def add_http_image_message(self, url: str):
        """Appends a user message with the image from the given url."""
        self.data.append(http_image_message(url))

    def add_user_prompt(self, name: str, **context: Any):
        """Appends a user message with the given pre-loaded prompt using jinja to render the context."""
        prompt = self._client.get_prompt(name)
        rendered_prompt = self._client.get_rendered_prompt(name, **context)
        message = _MetadataDict(user_message(rendered_prompt))
        message.prompt_metadata = {"name": name, "content": prompt, "context": context}
        self.data.append(message)  # pyright: ignore

    def add_system_prompt(self, name: str, **context: Any):
        """Appends a system message with the given pre-loaded prompt using jinja to render the context."""
        prompt = self._client.get_prompt(name)
        rendered_prompt = self._client.get_rendered_prompt(name, **context)
        message = _MetadataDict(system_message(rendered_prompt))
        message.prompt_metadata = {"name": name, "content": prompt, "context": context}
        self.data.append(message)  # pyright: ignore

    def add_assistant_prompt(self, name: str, **context: Any):
        """Appends a assistant message with the given pre-loaded prompt using jinja to render the context."""
        prompt = self._client.get_prompt(name)
        rendered_prompt = self._client.get_rendered_prompt(name, **context)
        message = _MetadataDict(assistant_message(rendered_prompt))
        message.prompt_metadata = {"name": name, "content": prompt, "context": context}
        self.data.append(message)  # pyright: ignore

    # Because the constructor has the client as an argument we have to redefine these methods from UserList
    def __getitem__(self, i):  # pyright: ignore
        if isinstance(i, slice):
            return self.__class__(self._client, self.data[i])
        else:
            return self.data[i]

    def __add__(self, other):
        if isinstance(other, UserList):
            return self.__class__(self._client, self.data + other.data)
        elif isinstance(other, type(self.data)):
            return self.__class__(self._client, self.data + other)
        return self.__class__(self._client, self.data + list(other))

    def __radd__(self, other):
        if isinstance(other, UserList):
            return self.__class__(self._client, other.data + self.data)
        elif isinstance(other, type(self.data)):
            return self.__class__(self._client, other + self.data)
        return self.__class__(self._client, list(other) + self.data)

    def __mul__(self, n):
        return self.__class__(self._client, self.data * n)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Callable, List

from dotenv import load_dotenv

from spice.errors import InvalidProviderError, NoAPIKeyError
from spice.wrapped_clients import WrappedAnthropicClient, WrappedAzureClient, WrappedClient, WrappedOpenAIClient

# Used to fetch a provider by name
providers: List[Provider] = []


@dataclass
class Provider:
    name: str
    get_client: Callable[[], WrappedClient]

    def __post_init__(self):
        providers.append(self)


def get_openai_client(cache=[]):
    if cache:
        return cache[0]

    base_url = os.getenv("OPENAI_API_BASE")
    key = os.getenv("OPENAI_API_KEY")
    if key is None:
        if base_url:
            # Using a dummy key for LiteLLM proxy setup where OPENAI_API_BASE is set but OPENAI_API_KEY is not required.
            key = "dummy_key"
        else:
            raise NoAPIKeyError("OPENAI_API_KEY not set")

    client = WrappedOpenAIClient(key, base_url)
    cache.append(client)
    return client


def get_azure_client(cache=[]):
    if cache:
        return cache[0]

    key = os.getenv("AZURE_OPENAI_KEY")
    endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
    if key is None:
        raise NoAPIKeyError("AZURE_OPENAI_KEY not set")
    if endpoint is None:
        raise NoAPIKeyError("AZURE_OPENAI_ENDPOINT not set")

    client = WrappedAzureClient(key, endpoint)
    cache.append(client)
    return client


def get_anthropic_client(cache=[]):
    if cache:
        return cache[0]

    key = os.getenv("ANTHROPIC_API_KEY")
    if key is None:
        raise NoAPIKeyError("ANTHROPIC_API_KEY not set")

    client = WrappedAnthropicClient(key)
    cache.append(client)
    return client


load_dotenv()

OPEN_AI = Provider("openai", get_openai_client)
AZURE = Provider("azure", get_azure_client)
ANTHROPIC = Provider("anthropic", get_anthropic_client)


def get_provider_from_name(provider_name: str) -> Provider:
    for provider in providers:
        if provider.name == provider_name:
            return provider

    raise InvalidProviderError(f"Unknown provider {provider_name}")



from __future__ import annotations

import dataclasses
from dataclasses import dataclass, field
from typing import List, Optional

from spice.providers import ANTHROPIC, OPEN_AI, Provider

# Used to fetch a model by name
models: List[Model] = []


# We give the superclass every single instance variable regardless of model type so that doing getattr isn't necessary when you have a generic Model
@dataclass
class Model:
    name: str
    """The model name to use when calling the provider."""

    provider: Provider | None
    """The provider this model belongs to. If this is None, a provider must always be passed alongside this model."""

    input_cost: Optional[int] = field(default=None)
    """This model's cost per input token in cents / million tokens, or cents / 100 seconds for transcription models."""

    output_cost: Optional[int] = field(default=None)
    """This model's cost per output token in cents / million tokens."""

    context_length: Optional[int] = field(default=None)
    """The maximum context length for this model."""

    def __post_init__(self):
        models.append(self)


@dataclass
class TextModel(Model):
    input_cost: Optional[int] = field(default=None)
    """This model's cost per input token in cents / million tokens."""

    output_cost: Optional[int] = field(default=None)
    """This model's cost per output token in cents / million tokens."""

    context_length: Optional[int] = field(default=None)
    """The maximum context length for this model."""


@dataclass
class VisionModel(TextModel):
    pass


@dataclass
class EmbeddingModel(Model):
    input_cost: Optional[int] = field(default=None)
    """This model's cost per input token in cents / million tokens."""

    context_length: Optional[int] = field(default=None)
    """The maximum context length for this model."""


@dataclass
class TranscriptionModel(Model):
    input_cost: Optional[int] = field(default=None)
    """This model's cost in cents / 100 seconds."""


@dataclass
class UnknownModel(TextModel, EmbeddingModel, TranscriptionModel):
    pass


GPT_4_TURBO = TextModel("gpt-4-turbo", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000)
"""Warning: This model always points to OpenAI's latest GPT-4-Turbo model, so the input and output costs may incorrect. We recommend using specific versions of GPT-4-Turbo instead."""

GPT_4_TURBO_2024_04_09 = TextModel(
    "gpt-4-turbo-2024-04-09", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000
)

GPT_4_TURBO_PREVIEW = TextModel(
    "gpt-4-turbo-preview", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000
)
GPT_4_0125_PREVIEW = TextModel("gpt-4-0125-preview", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000)
GPT_4_1106_PREVIEW = TextModel("gpt-4-1106-preview", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000)
GPT_4_VISION_PREVIEW = VisionModel(
    "gpt-4-vision-preview", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000
)
GPT_4_1106_VISION_PREVIEW = VisionModel(
    "gpt-4-1106-vision-preview", OPEN_AI, input_cost=1000, output_cost=3000, context_length=128000
)

GPT_4 = TextModel("gpt-4", OPEN_AI, input_cost=3000, output_cost=6000, context_length=8192)
"""Warning: This model always points to OpenAI's latest GPT-4 model, so the input and output costs may incorrect. We recommend using specific versions of GPT-4 instead."""

GPT_4_0613 = TextModel("gpt-4-0613", OPEN_AI, input_cost=3000, output_cost=6000, context_length=8192)

GPT_4_32K = TextModel("gpt-4-32k", OPEN_AI, input_cost=6000, output_cost=12000, context_length=32768)
"""Warning: This model always points to OpenAI's latest GPT-4-32k model, so the input and output costs may incorrect. We recommend using specific versions of GPT-4-32k instead."""

GPT_4_32K_0613 = TextModel("gpt-4-32k-0613", OPEN_AI, input_cost=6000, output_cost=12000, context_length=32768)

GPT_35_TURBO = TextModel("gpt-3.5-turbo", OPEN_AI, input_cost=50, output_cost=150, context_length=16385)
"""Warning: This model always points to OpenAI's latest GPT-3.5-Turbo model, so the input and output costs may incorrect. We recommend using specific versions of GPT-3.5-Turbo instead."""

GPT_35_TURBO_0125 = TextModel("gpt-3.5-turbo-0125", OPEN_AI, input_cost=50, output_cost=150, context_length=16385)
GPT_35_TURBO_1106 = TextModel("gpt-3.5-turbo-1106", OPEN_AI, input_cost=100, output_cost=200, context_length=16385)

GPT_35_TURBO_0613 = TextModel("gpt-3.5-turbo-0613", OPEN_AI, input_cost=150, output_cost=200, context_length=4096)
GPT_35_TURBO_16K_0613 = TextModel(
    "gpt-3.5-turbo-16k-0613", OPEN_AI, input_cost=300, output_cost=400, context_length=16385
)


CLAUDE_3_OPUS_20240229 = TextModel(
    "claude-3-opus-20240229", ANTHROPIC, input_cost=1500, output_cost=7500, context_length=200000
)
OPUS = CLAUDE_3_OPUS_20240229
CLAUDE_3_SONNET_20240229 = TextModel(
    "claude-3-sonnet-20240229", ANTHROPIC, input_cost=300, output_cost=1500, context_length=200000
)
SONNET = CLAUDE_3_SONNET_20240229
CLAUDE_3_HAIKU_20240307 = TextModel(
    "claude-3-haiku-20240307", ANTHROPIC, input_cost=25, output_cost=125, context_length=200000
)
HAIKU = CLAUDE_3_HAIKU_20240307


TEXT_EMBEDDING_3_LARGE = EmbeddingModel("text-embedding-3-large", OPEN_AI, input_cost=13, context_length=8191)
TEXT_EMBEDDING_3_SMALL = EmbeddingModel("text-embedding-3-small", OPEN_AI, input_cost=2, context_length=8191)
TEXT_EMBEDDING_ADA_002 = EmbeddingModel("text-embedding-ada-002", OPEN_AI, input_cost=10, context_length=8191)

WHISPER_1 = TranscriptionModel("whisper-1", OPEN_AI, input_cost=1)


def get_model_from_name(model_name: str) -> Model:
    # Search backwards; this way user defined models take priority
    for model in reversed(models):
        if model.name == model_name:
            return model

    if model_name.startswith("ft:"):
        adjusted_name = model_name.split(":")[1]
        for model in reversed(models):
            if model.name == adjusted_name:
                if "gpt-3.5-turbo" in model_name:
                    input_cost = 300
                    output_cost = 600
                elif "gpt-4" in model_name:
                    input_cost = 4500
                    output_cost = 9000
                else:
                    input_cost = None
                    output_cost = None
                model = dataclasses.replace(model, name=model_name, input_cost=input_cost, output_cost=output_cost)
                return model

    return UnknownModel(model_name, None)



from typing import Optional

from spice.models import EmbeddingModel, TextModel, TranscriptionModel


def text_request_cost(model: TextModel, input_tokens: int, output_tokens: int) -> Optional[float]:
    if model.input_cost is not None and model.output_cost is not None:
        return (model.input_cost * input_tokens + model.output_cost * output_tokens) / 1_000_000
    else:
        return None


def embeddings_request_cost(model: EmbeddingModel, input_tokens: int) -> Optional[float]:
    if model.input_cost is not None:
        return (model.input_cost * input_tokens) / 1_000_000
    else:
        return None


def transcription_request_cost(model: TranscriptionModel, input_length: float) -> Optional[float]:
    if model.input_cost is not None:
        return (model.input_cost * input_length) / 100
    else:
        return None


def print_stream(text: str) -> None:
    print(text, end="", flush=True)



import asyncio
import os
import sys
from typing import List

# Modify sys.path to ensure the script can run even when it's not part of the installed library.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from spice import Spice
from spice.spice_message import SpiceMessage


async def basic_example():
    client = Spice()

    messages: List[SpiceMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "list 5 random words"},
    ]
    response = await client.get_response(messages=messages, model="gpt-4-0125-preview")

    print(response.text)


async def streaming_example():
    # You can set a default model for the client instead of passing it with each call
    client = Spice(default_text_model="claude-3-opus-20240229")

    # You can easily load prompts from files, directories, or even urls.
    client.load_prompt("scripts/prompt.txt", name="my prompt")

    # Spice can also automatically render Jinja templates.
    messages: List[SpiceMessage] = [
        {"role": "system", "content": client.get_rendered_prompt("my prompt", assistant_name="Ryan Reynolds")},
        {"role": "user", "content": "list 5 random words"},
    ]
    stream = await client.stream_response(messages=messages)

    async for text in stream:
        print(text, end="", flush=True)
    # Retrieve the complete response from the stream
    response = await stream.complete_response()

    # Response always includes the final text, no need build it from the stream yourself
    print(response.text)

    # Response also includes helpful stats
    print(f"Took {response.total_time:.2f}s")
    print(f"Input/Output tokens: {response.input_tokens}/{response.output_tokens}")


async def multiple_providers_example():
    # Commonly used models and providers have premade constants
    from spice.models import GPT_4_0125_PREVIEW

    # Alias models for easy configuration, even mixing providers
    model_aliases = {
        "task1_model": GPT_4_0125_PREVIEW,
        "task2_model": "claude-3-opus-20240229",
        "task3_model": "claude-3-haiku-20240307",
    }

    client = Spice(model_aliases=model_aliases)

    messages: List[SpiceMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "list 5 random words"},
    ]
    responses = await asyncio.gather(
        client.get_response(messages=messages, model="task1_model"),
        client.get_response(messages=messages, model="task2_model"),
        client.get_response(messages=messages, model="task3_model"),
    )

    for i, response in enumerate(responses, 1):
        print(f"\nModel {i} response:")
        print(response.text)
        print(f"Characters per second: {response.characters_per_second:.2f}")
        if response.cost is not None:
            print(f"Cost: ${response.cost / 100:.4f}")

    # Spice also tracks the total cost over multiple models and providers
    print(f"Total Cost: ${client.total_cost / 100:.4f}")


async def azure_example():
    client = Spice()

    messages: List[SpiceMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "list 5 random words"},
    ]

    # To use Azure, specify the provider and the deployment model name
    response = await client.get_response(messages=messages, model="first-gpt35", provider="azure")
    print(response.text)

    # Alternatively, to make a model and it's provider known to Spice, create a custom Model object
    from spice.models import TextModel
    from spice.providers import AZURE

    AZURE_GPT = TextModel("first-gpt35", AZURE, context_length=16385)
    response = await client.get_response(messages=messages, model=AZURE_GPT)
    print(response.text)

    # Creating the model automatically registers it in Spice's model list, so listing the provider is no longer needed
    response = await client.get_response(messages=messages, model="first-gpt35")
    print(response.text)


async def vision_example():
    client = Spice()

    # Spice makes it easy to add images from files or the internet
    from spice import SpiceMessage, SpiceMessages
    from spice.models import CLAUDE_3_OPUS_20240229, GPT_4_1106_VISION_PREVIEW
    from spice.spice_message import file_image_message, user_message

    messages: List[SpiceMessage] = [user_message("What do you see?"), file_image_message("~/.mentat/picture.png")]
    response = await client.get_response(messages, GPT_4_1106_VISION_PREVIEW)
    print(response.text)

    # Alternatively, you can use the SpiceMessages wrapper to easily create your prompts
    spice_messages: SpiceMessages = SpiceMessages(client)
    spice_messages.add_user_message("What do you see?")
    spice_messages.add_file_image_message("~/.mentat/picture.png")
    response = await client.get_response(spice_messages, CLAUDE_3_OPUS_20240229)
    print(response.text)


async def embeddings_and_transcription_example():
    client = Spice()
    input_texts = ["Once upon a time...", "Cinderella"]

    # Spice can easily fetch embeddings and audio transcriptions
    from spice.models import TEXT_EMBEDDING_ADA_002, WHISPER_1

    embeddings = await client.get_embeddings(input_texts, TEXT_EMBEDDING_ADA_002)
    transcription = await client.get_transcription("~/.mentat/logs/audio/talk_transcription.wav", WHISPER_1)
    print(transcription.text)


async def run_all_examples():
    print("Running basic example:")
    await basic_example()
    print("\n\nRunning streaming example:")
    await streaming_example()
    print("\n\nRunning multiple providers example:")
    await multiple_providers_example()
    print("Running Azure example:")
    await azure_example()
    print("\n\nRunning vision example:")
    await vision_example()
    print("\n\nRunning embeddings and transcription example:")
    await embeddings_and_transcription_example()


if __name__ == "__main__":
    asyncio.run(run_all_examples())



from .spice import Spice, SpiceResponse, StreamingSpiceResponse, EmbeddingResponse, TranscriptionResponse  # noqa
from .spice_message import SpiceMessage, SpiceMessages  # noqa
from .utils import print_stream  # noqa



import pytest

from spice import Spice
from spice.models import HAIKU
from tests.conftest import WrappedTestClient


@pytest.mark.asyncio
async def test_get_response():
    client = WrappedTestClient(iter(["Hello, world!", "test"]))

    def equals_test(response):
        return response == "test"

    validator = equals_test

    spice = Spice()

    def return_wrapped_client(model, provider):
        return client

    spice._get_client = return_wrapped_client
    cache = ""

    def accumulator(text: str):
        nonlocal cache
        cache += text

    response = await spice.get_response(
        messages=[], model=HAIKU, validator=validator, streaming_callback=accumulator, retries=2
    )

    assert response.text == "test"
    assert response.retries == 1
    assert cache == "Hello, world!test"



class SpiceError(Exception):
    pass


class AuthenticationError(SpiceError):
    pass


class APIConnectionError(SpiceError):
    pass


class APIError(SpiceError):
    pass


class UnknownModelError(SpiceError):
    """Raised when an unknown model is passed in but no provider is given"""


class InvalidModelError(SpiceError):
    """Raised when a model is used in a situation it can't be used in; i.e., a text model passed into embeddings, or response format given to an Anthropic model."""


class InvalidProviderError(SpiceError):
    pass


class NoAPIKeyError(SpiceError):
    pass


class ImageError(SpiceError):
    pass
